@@ -3,7 +3,7 @@
 
 from snowflake.snowpark import functions as F
 from snowflake.snowpark import context
-from snowflake.snowpark.functions import call_builtin, col,lit, concat, coalesce, object_construct_keep_null
+from snowflake.snowpark.functions import call_builtin, col,lit, concat, coalesce, object_construct_keep_null, table_function, udf
 from snowflake.snowpark import DataFrame, Column
 from snowflake.snowpark.types import ArrayType, BooleanType
 from snowflake.snowpark._internal.type_utils import (
@@ -52,13 +52,6 @@ if not hasattr(F,"___extended"):
         # we add .* to the expression if needed
         return coalesce(call_builtin('regexp_substr',value,regexp,lit(1),lit(1),lit('e'),idx),lit(''))
 
-    def unix_timestamp(col):
-        return call_builtin("DATE_PART","epoch_second",col)
-
-    def from_unixtime(col):
-        col = _to_col_if_str(col,"from_unixtime")
-        return F.to_timestamp(col).alias('ts')
-
     def format_number(col,d):
         col = _to_col_if_str(col,"format_number")
         return F.to_varchar(col,'999,999,999,999,999.' + '0'*d)
@@ -100,7 +93,7 @@ if not hasattr(F,"___extended"):
             col_list.append(value)
         return object_construct(*col_list)
 
-    def array_distinct(col):
+    def _array_distinct(col):
         col = _to_col_if_str(col,"array_distinct")
         return F.call_builtin('array_distinct',col)
 
@@ -108,15 +101,15 @@ if not hasattr(F,"___extended"):
     def _array(*cols):
         return F.array_construct(*cols)
 
-    F._sort_array_function = None
+    F._sort_array_udf = None
     def _sort_array(col:ColumnOrName,asc:ColumnOrLiteral=True):
-        if not F._sort_array_function:
+        if not F._sort_array_udf:
             session = context.get_active_session()
             current_database = session.get_current_database()
             function_name =_generate_prefix("_sort_array_helper")
-            F._sort_array_function = f"{current_database}.public.{function_name}"
+            F._sort_array_udf = f"{current_database}.public.{function_name}"
             session.sql(f"""
-            create or replace temporary function {F._sort_array_function}(ARR ARRAY,ASC BOOLEAN) returns ARRAY
+            create or replace temporary function {F._sort_array_udf}(ARR ARRAY,ASC BOOLEAN) returns ARRAY
             language javascript as
             $$
             ARRLENGTH = ARR.length;
@@ -168,18 +161,18 @@ if not hasattr(F,"___extended"):
             var RES = new Array(ARRLENGTH-ARR.length).fill(null).concat(ARR);
             if (ASC) return RES; else return RES.reverse();
         $$;""").show()
-        return call_builtin(F._sort_array_function,col,asc)
+        return call_builtin(F._sort_array_udf,col,asc)
 
 
-    F._array_sort_function = None
+    F._array_sort_udf = None
     def _array_sort(col:ColumnOrName):
-        if not F._array_sort_function:
+        if not F._array_sort_udf:
             session = context.get_active_session()
             current_database = session.get_current_database()
             function_name =_generate_prefix("_array_sort_helper")
-            F._array_sort_function = f"{current_database}.public.{function_name}"
+            F._array_sort_udf = f"{current_database}.public.{function_name}"
             session.sql(f"""
-            create or replace temporary function {F._array_sort_function}(ARR ARRAY) returns ARRAY
+            create or replace temporary function {F._array_sort_udf}(ARR ARRAY) returns ARRAY
             language javascript as
             $$
             ARRLENGTH = ARR.length;
@@ -231,37 +224,37 @@ if not hasattr(F,"___extended"):
         var RES = ARR.concat(new Array(ARRLENGTH-ARR.length).fill(null));
         return RES;
         $$;""").show()
-        return call_builtin(F._array_sort_function,col)        
-    F._array_max_function = None
+        return call_builtin(F._array_sort_udf,col)        
+    F._array_max_udf = None
     def _array_max(col:ColumnOrName):
-        if not F._array_max_function:
+        if not F._array_max_udf:
             session = context.get_active_session()
             current_database = session.get_current_database()
             function_name =_generate_prefix("_array_max_function")
-            F._array_max_function = f"{current_database}.public.{function_name}"
+            F._array_max_udf = f"{current_database}.public.{function_name}"
             session.sql(f"""
-            create or replace temporary function {F._array_max_function}(ARR ARRAY) returns VARIANT
+            create or replace temporary function {F._array_max_udf}(ARR ARRAY) returns VARIANT
             language javascript as
             $$
             return Math.max(...ARR);
             $$
             """).show()
-        return call_builtin(F._array_max_function,col)
-    F._array_min_function = None
+        return call_builtin(F._array_max_udf,col)
+    F._array_min_udf = None
     def _array_min(col:ColumnOrName):
-        if not F._array_min_function:
+        if not F._array_min_udf:
             session = context.get_active_session()
             current_database = session.get_current_database()
-            function_name =_generate_prefix("_array_min_function")
-            F._array_min_function = f"{current_database}.public.{function_name}"
+            function_name =_generate_prefix("_array_min_udf")
+            F._array_min_udf = f"{current_database}.public.{function_name}"
             session.sql(f"""
-            create or replace temporary function {F._array_min_function}(ARR ARRAY) returns VARIANT
+            create or replace temporary function {F._array_min_udf}(ARR ARRAY) returns VARIANT
             language javascript as
             $$
             return Math.min(...ARR);
             $$
             """).show()
-        return call_builtin(F._array_min_function,col)
+        return call_builtin(F._array_min_udf,col)
 
     def _struct(*cols):
         new_cols = []
@@ -280,6 +273,52 @@ if not hasattr(F,"___extended"):
                 new_cols.append(c)
         return object_construct_keep_null(*new_cols)
 
+    F._array_flatten_udf = None
+    def _array_flatten(array):
+        if not F._array_flatten_udf:
+            @udf
+            def _array_flatten(array_in:list) -> list:
+                flat_list = []
+                for sublist in array_in:
+                    if type(sublist) == list:
+                        flat_list.extend(sublist)
+                    else:
+                        flat_list.append(sublist)              
+                return flat_list
+            F._array_flatten_udf = _array_flatten
+        array = _to_col_if_str(array, "array_flatten")
+        return F._array_flatten_udf(array)
+
+    F._array_zip_udfs = {}
+
+    def build_array_zip_ddl(nargs:int):
+        function_name = _generate_prefix(f"array_zip_{nargs}")
+        args          = ",".join([f"list{x} ARRAY" for x in range(1,nargs+1)])
+        args_names    = ",".join([f"list{x}"       for x in range(1,nargs+1)])
+        return function_name,f"""
+CREATE OR REPLACE TEMPORARY FUNCTION {function_name}({args})
+returns ARRAY language python runtime_version = '3.8'
+handler = 'zip_list'
+as
+$$
+def zip_list({args_names}):
+    return list(zip({args_names}))
+$$;"""
+
+    def _arrays_zip(*lists):
+        nargs = len(lists)
+        if nargs < 2:
+            raise Exception("At least two list are needed for array_zip")
+        if not str(nargs) in F._array_zip_udfs:
+            try:
+                function_name, udf_ddl = build_array_zip_ddl(nargs)
+                context.get_active_session().sql(udf_ddl).show()
+                F._array_zip_udfs[str(nargs)] = function_name
+            except Exception as e:
+                raise Exception(f"Could not register support udf for array_zip. Error: {e}")
+        list_cols = [_to_col_if_str(x, "array_zip") for x in lists]
+        return F.call_builtin(F._array_zip_udfs[str(nargs)],*list_cols)
+
     def _bround(col: Column, scale: int = 0): 
         power = pow(F.lit(10), F.lit(scale))
         elevatedColumn = F.when(F.lit(0) == F.lit(scale), col).otherwise(col * power)
@@ -297,53 +336,101 @@ if not hasattr(F,"___extended"):
     def is_not_a_regex(pattern):
         return not has_special_char(pattern)
 
-    F._split_regex = None
-    F.snowflake_split = F.split
-    def _regexp_split(value:ColumnOrName, pattern:ColumnOrLiteralStr, limit:ColumnOrLiteral = -1):        
+    F._split_regex_udf = None
+    def _regexp_split(value:ColumnOrName, pattern:ColumnOrLiteralStr, limit:int = -1):  
         value = _to_col_if_str(value,"split_regex")                
-        if not F._split_regex:            
-            session = context.get_active_session()
-            current_database = session.get_current_database()            
-            def split_regex_definition(value:str, pattern:str, limit:int)->str:                
-                if limit == 1:                    
-                    return '[\''+ value +'\']'
-                else:
-                    limit = limit - 1
-                if limit < 0:
-                    limit = 0                
-                return re.split(pattern,value,limit)            
-            F._split_regex = session.udf.register(split_regex_definition,is_permanent=False,overwrite=True)      
-        # Replace parenthesis because re.split adds what is inside them into the result list 
-        # while Pyspark doesn't take in count
-        pattern = pattern.replace('(','').replace(')','') 
         pattern_col = pattern        
         if isinstance(pattern, str):
             pattern_col = lit(pattern)        
         if limit < 0 and isinstance(pattern, str) and is_not_a_regex(pattern):
-            F.snowflake_split(value, pattern_col)        
-        if isinstance(limit, int):
-            limit = lit(limit)
-        return F._split_regex (value, pattern_col, limit)
+            return F.split(value, pattern_col)  
+                    
+        session = context.get_active_session()
+        current_database = session.get_current_database() 
+        function_name =_generate_prefix("_regex_split_helper")           
+        F._split_regex_udf = f"{current_database}.public.{function_name}"
 
-    F.array = _array
-    F.array_max = _array_max
-    F.array_min = _array_min
-    F.array_distinct = array_distinct
+        session.sql(f"""CREATE OR REPLACE FUNCTION {F._split_regex_udf} (input String, regex String, limit INT)
+RETURNS ARRAY
+LANGUAGE JAVA
+RUNTIME_VERSION = '11'
+PACKAGES = ('com.snowflake:snowpark:latest')
+HANDLER = 'MyJavaClass.regex_split_run' 
+AS
+$$
+import java.util.regex.Pattern;
+public class MyJavaClass {{
+    public String[] regex_split_run(String input,String regex, int limit) {{
+        Pattern pattern = Pattern.compile(regex);
+        return pattern.split(input, limit);
+    }}}}$$;""").show()
+        return call_builtin(F._split_regex_udf, value, pattern_col, limit)
+
+    def _explode(expr,outer=False,map=False,use_compat=False):
+        value_col = "explode"
+        if map:
+            key = "key"
+            value_col = "value"
+        else:
+            key = _generate_prefix("KEY")
+        seq = _generate_prefix("SEQ")
+        path = _generate_prefix("PATH")
+        index = _generate_prefix("INDEX")
+        this = _generate_prefix("THIS")
+        flatten = table_function("flatten")
+        explode_res = flatten(input=expr,outer=lit(outer)).alias(seq,key,path,index,value_col,this)
+        # we patch the alias, to simplify explode use case where only one column is used
+        if not map:
+            explode_res.alias_adjust = lambda alias1 : [seq,key,path,index,alias1,this] 
+        # post action to execute after join
+        def post_action(df):
+            drop_columns = [seq,path,index,this] if map else [seq,key,path,index,this]
+            df = df.drop(drop_columns)
+            if use_compat:
+                # in case we need backwards compatibility with spark behavior
+                df=df.with_column(value_col,
+                F.iff(F.cast(value_col,ArrayType()) == F.array_construct(),lit(None),F.cast(value_col,ArrayType())))
+            return df
+        explode_res.post_action = post_action
+        return explode_res
+
+    def _explode_outer(expr,map=False, use_compat=False):
+        return _explode(expr,outer=True,map=map,use_compat=use_compat)
+
+    F._map_values_udf = None
+    def _map_values(col:ColumnOrName):
+        col = _to_col_if_str(col,"map_values")
+        if not F._map_values_udf:
+            @udf(replace=True,is_permanent=False)
+            def map_values(obj:dict)->list:
+                return list(obj.values())
+            F._map_values_udf = map_values
+        return F._map_values_udf(col)
+
+
+
+
+
+
+    F.array          = _array
+    F.array_max      = _array_max
+    F.array_min      = _array_min
+    F.array_flatten  = _array_flatten
+    F.array_distinct = _array_distinct
+    F.array_sort     = _array_sort
+    F.arrays_zip     = _arrays_zip
+    F.bround         = _bround
+    F.create_map     = create_map
+    F.daydiff        = daydiff
+    F.date_add       = date_add
+    F.date_sub       = date_sub
+    F.explode        = _explode
+    F.explode_outer  = _explode_outer
+    F.format_number  = format_number
+    F.flatten        = _array_flatten
+    F.map_values     = _map_values
     F.regexp_extract = regexp_extract
-    F.create_map = create_map
-    F.unix_timestamp = unix_timestamp
-    F.from_unixtime = from_unixtime
-    F.format_number = format_number
-    F.reverse = reverse
-    F.daydiff = daydiff
-    F.date_add = date_add
-    F.date_sub = date_sub
-    F.asc  = lambda col: _to_col_if_str(col, "asc").asc()
-    F.desc = lambda col: _to_col_if_str(col, "desc").desc()
-    F.asc_nulls_first = lambda col: _to_col_if_str(col, "asc_nulls_first").asc()
-    F.desc_nulls_first = lambda col: _to_col_if_str(col, "desc_nulls_first").asc()
-    F.sort_array = _sort_array
-    F.array_sort = _array_sort
-    F.struct = _struct
-    F.bround = _bround
-    F.regexp_split = _regexp_split
\ No newline at end of file
+    F.regexp_split   = _regexp_split
+    F.reverse        = reverse
+    F.sort_array     = _sort_array
+    F.struct         = _struct
\ No newline at end of file
